Package: gcrma
Title: Background Adjustment Using Sequence Information
Version: 2.83.0
Author: Jean(ZHIJIN) Wu, Rafael Irizarry
	with contributions from
	James MacDonald <jmacdon@med.umich.edu>	
	Jeff Gentry
Description: Background adjustment using sequence information
Maintainer: Z. Wu <zwu@stat.brown.edu>
License: LGPL
Depends: R (>= 2.6.0), affy (>= 1.23.2), graphics, methods, stats,
        utils
Imports: Biobase, affy (>= 1.23.2), affyio (>= 1.13.3), XVector,
        Biostrings (>= 2.11.32), splines, BiocManager
Suggests: affydata, tools, splines, hgu95av2cdf, hgu95av2probe
biocViews: Microarray, OneChannel, Preprocessing
git_url: https://git.bioconductor.org/packages/gcrma
git_branch: devel
git_last_commit: 2a7d479
git_last_commit_date: 2025-10-29
Repository: Bioconductor 3.23
Date/Publication: 2025-10-31
NeedsCompilation: yes
Packaged: 2025-10-31 21:53:26 UTC; biocbuild

useDynLib(gcrma)

importMethodsFrom(affy, cdfName, indexProbes, intensity, mm, pm,
                  "pm<-", pmindex, probeNames)

importMethodsFrom(Biobase, exprs, "exprs<-", featureNames, pData)

importFrom(affy, AllButCelsForReadAffy, checkValidFilenames,
           cleancdfname, read.probematrix, rma, xy2indices)

importFrom(affyio, read.celfile.header)

importFrom(Biobase, testBioCConnection)

importFrom(BiocManager, repositories)

importFrom(XVector, subseq, "subseq<-")
importClassesFrom(Biostrings, "DNAStringSet")
importFrom(Biostrings, DNAStringSet, complement)
importMethodsFrom(Biostrings, "subseq<-", as.character)

importFrom(graphics, matplot)

importFrom(methods, new)

importFrom(splines, ns)

importFrom(stats, lm, loess, mad, median, predict, quantile, runmed)

importFrom(utils, available.packages, contrib.url, data,
           install.packages)

export(average.for.PAV, base.profiles, base.profiles.mm, base.profiles.nc, bg.adjust.affinities, bg.adjust.constant, bg.adjust.fullmodel, bg.adjust.gcrma, bg.adjust.mm, bg.adjust.optical, bg.parameters.ns, check.probes, compute.affinities, compute.affinities2, compute.affinities.local, compute.affinity.coef, fast.bkg, gcrma, gcrma.bg.transformation, gcrma.bg.transformation.fast, gcrma.engine, gcrma.engine2, getCDF, getProbePackage, just.gcrma, justGCRMA, left.sigma, mem.bkg, PAV, plotBaseProfiles,GSB.adj)

bg.adjust.optical <- function(abatch,minimum=1,verbose=TRUE){
  Index <- unlist(indexProbes(abatch,"both"))
  Index <- Index[!is.na(Index)]
  if(verbose) cat("Adjusting for optical effect")
  for(i in 1:length(abatch)){
    if(verbose) cat(".")
    exprs(abatch)[Index,i] <- exprs(abatch)[Index,i] -
      min(exprs(abatch)[Index,i],na.rm=TRUE) + minimum
  }
  if(verbose) cat("Done.\n")
  
  abatch
}


##########################################################################################
bg.adjust.fullmodel<- function(pms,mms,ncs=NULL,apm,amm,anc=NULL,index.affinities,k=6*fast+0.25*(1-fast),rho=.7,fast=FALSE){
  if(is.null(ncs)){
    parameters <- bg.parameters.ns(mms[index.affinities],amm,apm)
    mu.pm <- vector("numeric",length(pms))
    mu.mm <-  vector("numeric",length(pms))
    mu.pm[index.affinities] <- parameters$bg.mu2
    mu.mm[index.affinities] <- parameters$bg.mu
    sigma<- parameters$bg.sigma
  }
  
  else{
    parameters <- bg.parameters.ns(ncs,anc,apm,amm)
    mu.pm <- vector("numeric",length(pms))
    mu.mm <-  vector("numeric",length(pms))
    mu.pm[index.affinities] <- parameters$bg.mu2
    mu.mm[index.affinities] <- parameters$bg.mu3
    sigma<- parameters$bg.sigma
  }
  
  ##fill in the pms for which we dont have affinities
  if(length(index.affinities)<length(pms)){
    mu.pm[-index.affinities] <- median(mu.pm[index.affinities])
    mu.mm[-index.affinities] <- median(mu.mm[index.affinities])
  } 
  
  if(fast){
    ##this is the unbiased  
    bhat <- exp(mu.pm + rho*(log(mms)-mu.mm) + 1/2*(1 - rho^2)*sigma^2)
    var.y=exp(2*mu.pm+sigma^2)*(exp(sigma^2)-exp(sigma^2*rho^2))
    return(gcrma.bg.transformation.fast(pms,bhat,var.y,k=k))
  }
  else return(gcrma.bg.transformation(pms,mu.pm + rho*(log(mms)-mu.mm),sqrt(1 - rho^2)*sigma,k=k))
}

##########################################################################################
bg.adjust.affinities<- function(pms,ncs,apm,anc,index.affinities,k=6*fast+0.25*(1-fast),fast=FALSE,nomm=FALSE){
  if(!nomm) parameters <- bg.parameters.ns(ncs[index.affinities],anc,apm)
  else parameters <- bg.parameters.ns(ncs,anc,apm)
  mu.pm <- vector("numeric",length(pms))
  mu.pm[index.affinities] <- parameters$bg.mu2
  sigma<- parameters$bg.sigma
  ##fill in the pms for which we dont have affinities
  if(length(index.affinities)<length(pms)){
    mu.pm[-index.affinities] <- median(mu.pm[index.affinities])
  } 
  
  if(fast){
    ##this is the unbiased  
    bhat <- exp(mu.pm +  1/2*sigma^2)
    var.y=exp(2*mu.pm+sigma^2)*(exp(sigma^2)-1)
    return(gcrma.bg.transformation.fast(pms,bhat,var.y,k=k))
  }
  else return(gcrma.bg.transformation(pms,mu.pm,sigma,k=k))
} 
##########################################################################################
bg.adjust.mm <- function(pms,mms,k=6*fast+0.25*(1-fast),fast=FALSE){
  mu <- log(mms)
  Index <- which(pms<mms)
  sigma <- sqrt(mean((log(pms)[Index]-mu[Index])^2))

  bhat <- exp(mu+0.5*sigma^2)
  var.y <- exp(2*mu+sigma^2)*(exp(sigma^2)-1)

  if(fast) return(gcrma.bg.transformation.fast(pms,bhat,var.y,k=k))
  else return(gcrma.bg.transformation(pms,bhat,var.y,k=k))
}

##########################################################################################
bg.adjust.constant <- function(x,k=6*fast+0.25*(1-fast),Q=0.25,fast=FALSE){
  mu <- log(quantile(x,Q))
  sigma <- left.sigma(log(x),mu)
 
  if(fast){
     bhat <- exp(mu+1/2*sigma^2)
     var.y <- rep(exp(2*mu+sigma^2)*(exp(sigma^2)-1),length(x))
     return(gcrma.bg.transformation.fast(x,bhat,var.y,k=k))
   }
  else return(gcrma.bg.transformation(x,rep(mu,length(x)),sigma,k=k))
}
left.sigma <- function(x,mu) sqrt(mean((x[x<mu]-mu)^2))

compute.affinities2 <- function (cdfname, verbose = TRUE) 
{
    if (verbose) 
        cat("Computing affinities")
    data(affinity.spline.coefs)
    affinity.basis.matrix <- ns(1:25, df = length(affinity.spline.coefs)/3)
    cleancdf <- cleancdfname(cdfname, addcdf = FALSE)
    cdfpackagename <- paste(cleancdf, "cdf", sep = "")
    probepackagename <- paste(cleancdf, "probe", sep = "")
    getCDF(cdfpackagename)
    getProbePackage(probepackagename)
    p <- get(probepackagename)
    p <- check.probes(p, cdfname)
    affinity.info <- new("AffyBatch", cdfName = cdfname)
    my.xy2i <- get("xy2i", paste("package:", cdfpackagename, sep = ""))
    all <- my.xy2i(p$x,p$y)
    all25 <- all[which(nchar(p$sequence)==25)]
    pseq <- p$sequence[which(nchar(p$sequence)==25)]
    prlen <- unique(nchar(pseq))
    stopifnot(length(prlen) == 1 & prlen[1]==25)
    A13 <- sum(affinity.basis.matrix[13, ] * affinity.spline.coefs[1:5])
    T13 <- 0
    C13 <- sum(affinity.basis.matrix[13, ] * affinity.spline.coefs[6:10])
    G13 <- sum(affinity.basis.matrix[13, ] * affinity.spline.coefs[11:15])
    if (verbose) 
        cat(".")
    apm <- vector("numeric", length(pseq))
    for (i in seq(along = apm)) {
        charMtrx <- .Call("gcrma_getSeq", pseq[i], PACKAGE = "gcrma")
        A <- cbind(charMtrx[1, ] %*% affinity.basis.matrix, charMtrx[2, 
            ] %*% affinity.basis.matrix, charMtrx[3, ] %*% affinity.basis.matrix)
        apm[i] <- A %*% affinity.spline.coefs
      }
    tmp.exprs = matrix(NA, nrow = max(all),ncol = 1)
    tmp.exprs[all25] = apm
    
    exprs(affinity.info) = tmp.exprs
    if (verbose) 
        cat("Done.\n")
    return(affinity.info)
}

complementSeq <- function(seq,start=1,stop=0){
  if (is.numeric(stop) && length(stop) == 1L && !is.na(stop) && stop == 0)
    stop <- -1
  seq <- DNAStringSet(seq)
  subseq(seq, start=start, end=stop) <-
    complement(subseq(seq, start=start, end=stop))
  as.character(seq)
}

# compute affinity.spline.coefs from MMs
base.profiles.mm <- function(object,verbose=TRUE){
  
  cleancdf <- cleancdfname(cdfName(object),addcdf=FALSE)
  ##  object <- bg.adjust.optical(object) #should already been done
  cdfpackagename <- paste(cleancdf,"cdf",sep="")
  #myxy2i <- get("xy2i",paste("package:",cdfpackagename,sep=""))
  probepackagename <- paste(cleancdf,"probe",sep="")
  getCDF(cdfpackagename)
  getProbePackage(probepackagename)
  p <- get(probepackagename)
  seqs=p$seq#the PM sequences
  seqs=complementSeq(seqs, start=13, stop=13)#the MM sequences

  pmIndex <-  unlist(indexProbes(object,"pm"))
  mmIndex <-  unlist(indexProbes(object,"mm"))
  subIndex <- match(xy2indices(p$x,p$y, cdf=cdfpackagename),pmIndex)
  bgy <- as.matrix(intensity(object)[mmIndex[subIndex],])
  affinity.spline.coefs <- base.profiles(bgy,seqs)
  affinity.spline.coefs
}
# compute affinity.spline.coefs from a vector of negative control probes

base.profiles.nc <- function(object, NCprobe,verbose=TRUE){
  cleancdf  <- cleancdfname(cdfName(object),addcdf=FALSE)
  ##object <- bg.adjust.optical(object)
  cdfpackagename <- paste(cleancdf,"cdf",sep="")
  #myxy2i <- get("xy2i",paste("package:",cdfpackagename,sep=""))
  probepackagename <- paste(cleancdf,"probe",sep="")
  getCDF(cdfpackagename)
  getProbePackage(probepackagename)
  p <- get(probepackagename)
  PMseq=p$seq#the PM sequences
  MMseq=complementSeq(PMseq, start=13, stop=13)#the MM sequences

  pmIndex <-  unlist(indexProbes(object,"pm"))
  mmIndex <-  unlist(indexProbes(object,"mm"))
  subIndex1 <- match(NCprobe,c(xy2indices(p$x,p$y, cdf=cdfpackagename),
                               xy2indices(p$x,p$y+1, cdf=cdfpackagename)))
  seqs <-c(PMseq,MMseq)[subIndex1[!is.na(subIndex1)]]

  bgy <- as.matrix(intensity(object)[NCprobe,])
  if(length(seqs)<length(NCprobe)){
    cat("\nNote: some of your negative control probes do not have sequence information\n")
    subIndex2 <- match(c(xy2indices(p$x,p$y, cdf=cdfpackagename),
                         xy2indices(p$x,p$y+1, cdf=cdfpackagename)),NCprobe)
    subIndex2 <- subIndex2[!is.na(subIndex2)]
    bgy <- bgy[!is.na(subIndex1),]}

  affinity.spline.coefs <- base.profiles(bgy,seqs)
  affinity.spline.coefs
}

##seqs: probe sequences of those used to estimate base profiles
##intensities: observed intensity on these probes, a vector or a matrix
base.profiles <- function(intensities,seqs){
  prlen <- unique(nchar(seqs))
  stopifnot(length(prlen)==1)
  mS <- .Call("gcrma_getSeq2",paste(seqs,collapse=""),length(seqs),PACKAGE="gcrma")

  ##Create a basis matrix
  B <- ns(1:25,df=5)
  ##Create the prediction matrix in the model
  X <- cbind(mS[,1:25]%*%B,mS[,26:50]%*%B,mS[,51:75]%*%B)
  
  Y=as.matrix(intensities)
  affinity.spline.coefs <- apply(Y,2,function(y)
                                 lm(log2(y) ~ X)$coef[-1])
  rownames(affinity.spline.coefs) <- c(rep("A",5),rep("C",5),rep("G",5))
  affinity.spline.coefs
}

plotBaseProfiles <- function(affinity.spline.coefs,prlen=25){
  if(!is.vector(affinity.spline.coefs)) stop("affinity.spline.coefs must be a vector")
  P <- length(affinity.spline.coefs)/3 ##number of parameters per letter
  B <- ns(1:25,df=5)
  coeff <- matrix(0, nrow=prlen, ncol=4)
  colnames(coeff) <- c("A", "C", "G", "T")
  for (i in 0:2)
    coeff[, i+1] <- B%*%affinity.spline.coefs[i*P+(1:P)]
  coeff <- coeff-rowSums(coeff[,1:3])/4
  matplot(coeff,  type="b", col=1:4, lwd=2, pch=colnames(coeff))
}

##this is almost the same as compute.affinities
##except that affinity.spline.coefs is not loaded with data()
##but computed using the data provided by the user
compute.affinities.local <- function(object,Array=1,NCprobe=NULL,verbose=TRUE,
                                     affinity.spline.coefs=NULL,optical.correct=TRUE){
  if(optical.correct) object <- bg.adjust.optical(object)
  if(is.null(affinity.spline.coefs))
    affinity.spline.coefs <- compute.affinity.coef(object,Array,NCprobe,verbose=TRUE)
  if(verbose) cat("Computing affinities")
  affinity.basis.matrix <- ns(1:25,df=nrow(as.matrix(affinity.spline.coefs))/3)

  cdfname <- cdfName(object)
  cleancdf <- cleancdfname(cdfname,addcdf=FALSE)
  cdfpackagename <- paste(cleancdf,"cdf",sep="")
  probepackagename <- paste(cleancdf,"probe",sep="")
  
  getCDF(cdfpackagename)
  getProbePackage(probepackagename)
  p <- get(probepackagename)
  
  p <- check.probes(p, cdfname)
  
  prlen <- unique(nchar(p$sequence))
  stopifnot(length(prlen)==1)
  
  A13 <- sum(affinity.basis.matrix[13,]*affinity.spline.coefs[1:5])
  T13 <- 0
  C13 <- sum(affinity.basis.matrix[13,]*affinity.spline.coefs[6:10])
  G13 <- sum(affinity.basis.matrix[13,]*affinity.spline.coefs[11:15])
  
  APM=AMM=matrix(NA,length(p$sequence),length(object))
  for(K in 1:ncol(APM)){
    if(verbose) cat(".")
    apm <- vector("numeric",length(p$sequence))
    amm <- vector("numeric",length(p$sequence))
    
    for(i in seq(along=apm)) {
      charMtrx <- .Call("gcrma_getSeq", p$sequence[i],
                        PACKAGE="gcrma")
      A <- cbind(charMtrx[1,] %*% affinity.basis.matrix,
                 charMtrx[2,] %*% affinity.basis.matrix,
                 charMtrx[3,] %*% affinity.basis.matrix)
      
      apm[i] <- A %*% affinity.spline.coefs[,K]
      
      if (charMtrx[1,13] == 1) {
        amm[i] <- apm[i] + T13 - A13
      }
      else {
        if (charMtrx[4,13] == 1) {
          amm[i] <- apm[i] + A13 - T13
        }
        else{
          if (charMtrx[3,13]) {
            amm[i] <- apm[i] + C13 - G13
          }
          else {
            amm[i] <- apm[i] + G13 - C13
          }
        }
      }
    }
    APM[,K]=apm;AMM[,K]=amm
  }
  ##put it in an affybatch
  #tmp <- get("xy2i",paste("package:",cdfpackagename,sep=""))
  affinity.info <- new("AffyBatch",cdfName=cdfname)
  pmIndex <-  unlist(indexProbes(affinity.info,"pm"))
  mmIndex <-  unlist(indexProbes(affinity.info,"mm"))
  subIndex <- match(xy2indices(p$x,p$y, cdf=cdfpackagename),pmIndex)
  tmp.exprs=matrix(NA,nrow=max(cbind(pmIndex,mmIndex)),ncol=ncol(APM))
  tmp.exprs[pmIndex[subIndex],]=APM
  if(!is.null(amm)){ tmp.exprs[mmIndex[subIndex],]=AMM }
  exprs(affinity.info)=tmp.exprs
  if(verbose) cat("Done.\n")
  return(affinity.info)
}

compute.affinity.coef <- function(object,Array,NCprobe,verbose=TRUE){
  if(verbose) cat("Computing base-position profiles for probe affinities")
  if(is.null(NCprobe)){ #use MMs
    if(is.null(Array))  #"use all arrays"
      affinity.spline.coefs <- base.profiles.mm(object,verbose)
    else if(is.numeric(Array) & length(Array)==1) #use one Array
      affinity.spline.coefs <- base.profiles.mm(object[,Array],verbose)
    else stop("\nParameter Array should be either NULL(use all arrays)
 or a number indexing the array to be used\n")
  }

  else if(is.numeric(NCprobe)){ #"use NCs from each array"
    if(is.null(Array))  #"use all arrays"
      affinity.spline.coefs <- base.profiles.nc(object,NCprobe,verbose)
    else if(is.numeric(Array) & length(Array)==1) #use one Array
      affinity.spline.coefs <- base.profiles.nc(object[,Array],NCprobe,verbose)
    else stop("\nParameter Array should be either NULL(use all arrays)
 or a number indexing the array to be used\n")
  }

  else
    stop("\nNCprobe should be either NULL (use MMs)
or a vector of probe indexes for negative control probes\n")
  if(verbose) cat("Done.\n")
  affinity.spline.coefs
}

##default, one array affinity.info
gcrma.engine <- function(pms,mms,ncs=NULL,
                         pm.affinities=NULL,mm.affinities=NULL,anc=NULL,
                         type=c("fullmodel","affinities","mm","constant"),
                         k=6*fast+0.5*(1-fast),
                         stretch=1.15*fast+1*(1-fast),correction=1,GSB.adjust=TRUE,rho=0.7,
                         verbose=TRUE,fast=FALSE){
  
  if(!is.null(pm.affinities)){
    index.affinities <- which(!is.na(pm.affinities))
    pm.affinities <- pm.affinities[index.affinities]
    if(!is.null(mm.affinities)){
      mm.affinities <- mm.affinities[index.affinities]
    }
  }
  
###get GSB.adjust parameters
  if(GSB.adjust &(type=="fullmodel" | type=="affinities")){
    set.seed(1)
      Subset <- sample(1:length(as.matrix(pms)[index.affinities,]),25000)
    y <- log2(pms)[index.affinities,][Subset]
    Subset <- (Subset-1)%%nrow(as.matrix(pms)[index.affinities,])+1
    x <- pm.affinities[Subset]
    fit1 <- lm(y~x)$coef
  }
  
  
  if(verbose) cat("Adjusting for non-specific binding")
  for(i in 1:ncol(pms)){
    if(verbose) cat(".")
    if(type=="fullmodel") {
      pms[,i] <- bg.adjust.fullmodel(pms[,i],mms[,i],ncs=ncs[,i],
                                     apm=pm.affinities,amm=mm.affinities,anc=anc,
                                     index.affinities,k=k,
                                     rho=rho,fast=fast)
      if(GSB.adjust) 
        pms[,i] <- GSB.adj(Yin=pms[,i],subset=index.affinities,aff=pm.affinities,fit1=fit1,k=k)
    }
    if(type=="affinities"){
      if(is.null(ncs))
        pms[,i] <-  bg.adjust.affinities(pms[,i],ncs=mms[,i],apm=pm.affinities,anc=mm.affinities,
                                         index.affinities,k=k,fast=fast)
      else
        pms[,i] <- bg.adjust.affinities(pms[,i],ncs=ncs[,i],apm=pm.affinities,anc=anc,
                                        index.affinities,k=k,fast=fast,nomm=TRUE)
      if(GSB.adjust)
        pms[,i] <- GSB.adj(Yin=pms[,i],subset=index.affinities,aff=pm.affinities,fit1=fit1,k=k)
    }
    if(type=="mm") pms[,i] <- bg.adjust.mm(pms[,i],correction*mms[,i],k=k,fast=fast)
    if(type=="constant"){
      pms[,i] <- bg.adjust.constant(pms[,i],k=k,Q=correction*mean(pms[,i]<mms[,i]),fast=fast)
    }
    if(stretch!=1){
      mu <- mean(log(pms[,i]))
      pms[,i] <- exp(mu + stretch*(log(pms[,i])-mu))
    }
  }

  if(verbose) cat("Done.\n")
  return(pms)
}


GSB.adj <- function(Yin,subset,aff,fit1,k=k){ #subset are the index of Yin with the affinities aff
    y0=Yin[subset]
    y0.adj=k+ 2^(-fit1[2]*(aff-mean(aff)))*(y0-k)
    Yin[subset]=y0.adj
    Yin
     }

compute.affinities <- function(cdfname,verbose=TRUE){
  if(verbose) cat("Computing affinities")   
  data(affinity.spline.coefs, package = "gcrma") ###needs to change to data(something)
  affinity.basis.matrix <- ns(1:25,df=length(affinity.spline.coefs)/3)
  
  cleancdf <- cleancdfname(cdfname,addcdf=FALSE)
  cdfpackagename <- paste(cleancdf,"cdf",sep="")
  probepackagename <- paste(cleancdf,"probe",sep="")
  
  getCDF(cdfpackagename)
  getProbePackage(probepackagename)
  p <- get(probepackagename)
  
  p <- check.probes(p, cdfname)
  
  prlen <- unique(nchar(p$sequence))
  stopifnot(length(prlen)==1)
  
  A13 <- sum(affinity.basis.matrix[13,]*affinity.spline.coefs[1:5])
  T13 <- 0
  C13 <- sum(affinity.basis.matrix[13,]*affinity.spline.coefs[6:10])
  G13 <- sum(affinity.basis.matrix[13,]*affinity.spline.coefs[11:15])
  
  if(verbose) cat(".")
  
  apm <- vector("numeric",length(p$sequence))
  amm <- vector("numeric",length(p$sequence))
  
  for(i in seq(along=apm)) {
    charMtrx <- .Call("gcrma_getSeq", p$sequence[i],
                      PACKAGE="gcrma")
    A <- cbind(charMtrx[1,] %*% affinity.basis.matrix,
               charMtrx[2,] %*% affinity.basis.matrix,
               charMtrx[3,] %*% affinity.basis.matrix)
    
    apm[i] <- A %*% affinity.spline.coefs
    
    if (charMtrx[1,13] == 1) {
      amm[i] <- apm[i] + T13 - A13
    }
    else {
      if (charMtrx[4,13] == 1) {
        amm[i] <- apm[i] + A13 - T13
      }
      else{
        if (charMtrx[3,13]) {
          amm[i] <- apm[i] + C13 - G13
        }
        else {
          amm[i] <- apm[i] + G13 - C13
        }
      }
    }
  }
  
  ##put it in an affybatch
  #tmp <- get("xy2i",paste("package:",cdfpackagename,sep=""))
  affinity.info <- new("AffyBatch",cdfName=cdfname)
  pmIndex <-  unlist(indexProbes(affinity.info,"pm"))
  mmIndex <-  unlist(indexProbes(affinity.info,"mm"))
  subIndex <- match(xy2indices(p$x,p$y, cdf=cdfpackagename),pmIndex)
  tmp.exprs=matrix(NA,nrow=max(cbind(pmIndex,mmIndex)),ncol=1)
  tmp.exprs[pmIndex[subIndex]]=apm
  if(!is.null(amm)){ tmp.exprs[mmIndex[subIndex]]=amm }
  exprs(affinity.info)=tmp.exprs
  if(verbose) cat("Done.\n")
  return(affinity.info)
}

check.probes <- function(probepackage, cdfname){
  cdfnames <- names(pmindex(new("AffyBatch", cdfName=cdfname)))
  ppnames <- as.character(probepackage$Probe.Set.Name)
  
  if (sum(!(ppnames %in% cdfnames)) != 0){
    Index <- ppnames %in% cdfnames
    probepackage <- probepackage[Index,]
  }
  return(probepackage)
}

compute.affinities.nomm <- function (cdfname, verbose = TRUE) 
{
    if (verbose) 
        cat("Computing affinities")
    data(affinity.spline.coefs)
    affinity.basis.matrix <- ns(1:25, df = length(affinity.spline.coefs)/3)
    cleancdf <- cleancdfname(cdfname, addcdf = FALSE)
    cdfpackagename <- paste(cleancdf, "cdf", sep = "")
    probepackagename <- paste(cleancdf, "probe", sep = "")
    getCDF(cdfpackagename)
    getProbePackage(probepackagename)
    p <- get(probepackagename)
    p <- check.probes(p, cdfname)
    affinity.info <- new("AffyBatch", cdfName = cdfname)
   pmIndex <- unlist(indexProbes(affinity.info, "pm"))
    #pm probes with seq
    my.xy2i <- get("xy2i", paste("package:", cdfpackagename, sep = ""))
    index1 <- match(pmIndex,my.xy2i(p$x,p$y))
    subIndex <- index1[!is.na(index1)]
    pmseq <- p$sequence[subIndex]
    prlen <- unique(nchar(pmseq))
    stopifnot(length(prlen) == 1 & prlen[1]==25)
    A13 <- sum(affinity.basis.matrix[13, ] * affinity.spline.coefs[1:5])
    T13 <- 0
    C13 <- sum(affinity.basis.matrix[13, ] * affinity.spline.coefs[6:10])
    G13 <- sum(affinity.basis.matrix[13, ] * affinity.spline.coefs[11:15])
    if (verbose) 
        cat(".")
    apm <- vector("numeric", length(pmseq))
    for (i in seq(along = apm)) {
        charMtrx <- .Call("gcrma_getSeq", pmseq[i], PACKAGE = "gcrma")
        A <- cbind(charMtrx[1, ] %*% affinity.basis.matrix, charMtrx[2, 
            ] %*% affinity.basis.matrix, charMtrx[3, ] %*% affinity.basis.matrix)
        apm[i] <- A %*% affinity.spline.coefs
      }
    tmp <- get("xy2i", paste("package:", cdfpackagename, sep = ""))
    tmp.exprs = matrix(NA, nrow = max(pmIndex),ncol = 1)
    tmp.exprs[pmIndex[match(subIndex,pmIndex)]] = apm
    exprs(affinity.info) = tmp.exprs
    if (verbose) 
        cat("Done.\n")
    return(affinity.info)
}

getCDF <- function(cdfname, lib=.libPaths()[1], verbose=TRUE){

  ## 1-23-2006 JWM: removed dependency on reposTools
  
  options(show.error.messages = FALSE)
  attempt <- try(do.call(library, list(cdfname, lib.loc = lib)))
  options(show.error.messages = TRUE)
  if (inherits(attempt, "try-error")){
    
    if(testBioCConnection()){
      if(verbose)
        print("Checking to see if your internet connection works...")
      ## Check for file permissions
      if(file.access(lib, mode = 0) < 0) stop(paste("Directory", lib,"does not",
                            "seem to exist.\n", "Please check your 'lib' parameter",
                            "and try again."))
      
      if(file.access(lib, mode = 2) < 0) stop(paste("You do not have write access to", lib,
                            "\nPlease check your permissions or provide",
                            "a different 'lib' parameter."))
      
      biocContribUrl <- sapply(repositories(), contrib.url)
      biocPkgs <- available.packages(biocContribUrl)
      if(!cdfname %in% biocPkgs[,"Package"]){
        if(verbose)
          print(paste("Environment",cdfname,
                      "was not found in the Bioconductor",
                      "repository."))
        return(list(paste("Bioconductor -",cdfname,"not available")))
      }else{
        install.packages(cdfname, lib=lib,
                         repos=repositories(),
                         dependencies=TRUE)
      }
    }else{
      stop(paste("The current operation could not access",
                 "the Bioconductor repository. Please",
                 "check your internet connection, and",
                 "report further problems to",
                 "bioconductor@stat.math.ethz.ch"))
    }
  }
  ## Now load the library
  do.call(library, list(cdfname, lib.loc=lib))
  ##Check that library is loaded
  if(!cdfname %in% .packages()) stop(paste("The package", cdfname,
                                           "could not be loaded."))
}


getProbePackage <- function(probepackage, lib=.libPaths()[1], verbose=TRUE){

  ## 1-23-2006 JWM: Removed dependency on reposTools
  options(show.error.messages = FALSE)
  attempt <- try(do.call(library, list(probepackage, lib.loc = lib)))
  options(show.error.messages = TRUE)
  if (inherits(attempt, "try-error")){
    
    if(testBioCConnection()){
      if(verbose)
        print("Checking to see if your internet connection works...")
      ## Check for file permissions
      if(file.access(lib, mode = 0) < 0) stop(paste("Directory", lib,"does not",
                            "seem to exist.\n", "Please check your 'lib' parameter",
                            "and try again."))
      
      if(file.access(lib, mode = 2) < 0) stop(paste("You do not have write access to", lib,
                            "\nPlease check your permissions or provide",
                            "a different 'lib' parameter."))
      
      biocContribUrl <- sapply(repositories(), contrib.url)
      biocPkgs <- available.packages(biocContribUrl)
      if(!probepackage %in% biocPkgs[,"Package"]){
        if(verbose)
          print(paste("Environment",probepackage,
                      "was not found in the Bioconductor",
                      "repository."))
        return(list(paste("Bioconductor -",probepackage,"not available")))
      }else{
        install.packages(probepackage, lib=lib,
                         repos=repositories(),
                         dependencies=TRUE)
      }
    }else{
      stop(paste("The current operation could not access",
                 "the Bioconductor repository. Please",
                 "check your internet connection, and",
                 "report further problems to",
                 "bioconductor@stat.math.ethz.ch"))
    }
  }
  
  ## Now load the library
  do.call(library, list(probepackage, lib.loc=lib))
  ##Check that library is loaded
  if(!probepackage %in% .packages()) stop(paste("The package", probepackage,
                                                "could not be loaded."))
}

average.for.PAV <-function(y, wt = rep(1, length(y)))
{
  ## compute a weighted average of a vector, y
  if(any(is.na(wt))) stop("NA's not allowed for wt")
  if(any(wt < 0))
    stop("wt must be a vector of NON-NEGATIVE weights")
  if(length(wt) != length(y)) stop("y and wt must be vectors of the same length")
  ## if any observations have Infinite weight, return the simple
  ## (unweighted) average of only those observations (giving no
  ## weight to observations with finite weight)
  if(any(wt == Inf)) {
    wt[wt < Inf] <- 0
    wt[wt == Inf] <- 1
  }
  ## if all weights are zero, return the simple (unweighted)
  ## average of y
  if(sum(wt) == 0)
    wt <- rep(1, length(wt))
  return(sum((y * wt)/sum(wt)))
}

PAV <- function(y, wt = rep(1,length(y)))
{
  ## This is a modification of Derick's PAV program
  ##
  ## (Weighted) Pool-Adjacent-Violators (PAV) algorithm
  ## for non-parametric monotonic (decreasing) regression of y on x
  n <- length(y)
  if(n != length(wt))
    stop("y, and wt must be vectors of equal length")
  yhat <- y       # initialize while loop
  j <- count <- 1
  k <- 2
  support <- vector("numeric", n)
  support[count] <- j
  while(k <= n) {
    while(yhat[j] < yhat[k]) {
      yhat[j:k] <- average.for.PAV(y[j:k], wt[j:k])
      if(yhat[support[count]] < yhat[k]) {
        j <- support[count]
        if(count > 1)
          count <- count - 1
      }
      else {
        k <- ifelse(k == n, k, k + 1)
      }
    }
    
    count <- count + 1
    support[count] <- j
    j <- k
    k <- k + 1
  }
  return(list(y = yhat, wt))
}

gcrma.bg.transformation.fast <- function(x,bhat,var.y,k){
  
  x <- x - bhat ##this is an unbiased estimate
  Index <- x > 0
  x[!Index] <- 0 ##trick to not get -Inf in log
  
  ##alpha is the weith used to create weighted estimate
  ##its basically  for smoothing edge
  alpha <- rep(0,length(x))
  xplusk <- x[Index]+k  #x plus k
  logxplusk <- log(xplusk)
  alpha[Index] <- (logxplusk - log(k))*(logxplusk - log(k) + k/xplusk) /
    ( ( logxplusk - log(k) + k/xplusk)^2 + var.y[Index]/xplusk^2)
  
  y <- exp(alpha*log(x+k) + (1-alpha)*log(k))
}

### A user friendly wrapper for just.gcrma
justGCRMA <- function(..., filenames=character(0),
                     widget=getOption("BioC")$affy$use.widgets,
                     compress=getOption("BioC")$affy$compress.cel,
                     celfile.path=getwd(),
                     sampleNames=NULL,
                     phenoData=NULL,
                     description=NULL,
                     notes="",
                     normalize=TRUE, 
                     bgversion=2, affinity.info=NULL,
                     type=c("fullmodel","affinities","mm","constant"),
                     k=6*fast+0.5*(1-fast), stretch=1.15*fast+1*(1-fast),
                     correction=1, rho=0.7, optical.correct=TRUE,
                     verbose=TRUE, fast=TRUE, minimum=1, optimize.by = c("speed","memory"),
                      cdfname = NULL, read.verbose = FALSE){

   l <- AllButCelsForReadAffy(..., filenames=filenames,
                             widget=widget,
                             celfile.path=celfile.path,
                             sampleNames=sampleNames,
                             phenoData=phenoData,
                             description=description)
  

  ##and now we are ready to read cel files
  return(just.gcrma(filenames=l$filenames,
                    phenoData=l$phenoData,
                    description=l$description,
                    notes=notes,
                    compress=compress,
                    verbose=verbose,
                    normalize=normalize,
                    bgversion=bgversion,
                    affinity.info=affinity.info,
                    type=type, k=k, stretch=stretch,
                    correction=correction, rho=rho,
                    optical.correct=optical.correct,
                    fast=fast, minimum=minimum,
                    optimize.by=optimize.by,
                    cdfname = cdfname,
                    read.verbose = read.verbose))

}


just.gcrma <- function(..., filenames=character(0),
                       phenoData=new("AnnotatedDataFrame"),
                       description=NULL,
                       notes="", compress=getOption("BioC")$affy$compress.cel,
                       normalize=TRUE, bgversion=2, affinity.info=NULL,
                       type=c("fullmodel","affinities","mm","constant"),
                       k=6*fast+0.5*(1-fast), stretch=1.15*fast+1*(1-fast),
                       correction=1, rho=0.7, optical.correct=TRUE,
                       verbose=TRUE, fast=TRUE, minimum=1, optimize.by = c("speed","memory"),
                       cdfname = NULL, read.verbose = FALSE) {

  auxnames <- unlist(list(...))
  filenames <- c(filenames, auxnames)

  checkValidFilenames(filenames)
  
  n <- length(filenames)
  
  ## error if no file name !
  if (n == 0)
    stop("No file name given !")
  
  pdata <- pData(phenoData)
  ##try to read sample names from phenoData. if not there use CEL filenames
  if(dim(pdata)[1]!=n){#if empty pdata filename are samplenames
    #warning("Incompatible phenoData object. Created a new one.\n")
    
    samplenames <- gsub("^/?([^/]*/)*", "", unlist(filenames))
    pdata <- data.frame(sample=1:n,row.names=samplenames)
    phenoData <- new("AnnotatedDataFrame",
                     data = pdata,
                     varMetadata = data.frame(
                     labelDescription = "arbitrary numbering",
                     row.names = "sample"))
  }
  else samplenames <- rownames(pdata)
  

  if (is.null(description))
  {
    description <- new("MIAME")
    description@preprocessing$filenames <- filenames
    description@preprocessing$affyversion <- library(help=affy)$info[[2]][[2]][2]
  }

  ## get information from cdf environment

  headdetails <- read.celfile.header(filenames[[1]])
  dim.intensity <- headdetails[[2]]
  if(is.null(cdfname))
    cdfName <- headdetails[[1]]
  else
    cdfName <- cdfname
  
  type <- match.arg(type)

  pmonly <- (type=="affinities"|type=="constant")
  needaff <- (type=="fullmodel"|type=="affinities")

  if( needaff ){
    if(is.null (affinity.info)){
      affinity.info <- compute.affinities(cdfName,verbose=verbose)
    }
      
    pm.affinities <- pm(affinity.info)
    mm.affinities <- mm(affinity.info)
    
    index.affinities <- which(!is.na(pm.affinities))
    
    pm.affinities <- pm.affinities[index.affinities]
    mm.affinities <- mm.affinities[index.affinities]
    
    ##Recover memory
    rm(affinity.info)
    gc()
    
  }

  speed <- match.arg(optimize.by)
  if(speed == "speed"){
    pms <- fast.bkg(filenames = filenames, pm.affinities = pm.affinities,
                    mm.affinities = mm.affinities, index.affinities = index.affinities,
                    type = type, minimum = minimum, optical.correct = optical.correct,
                    verbose = verbose, k = k, rho = rho, correction = correction,
                    stretch = stretch, fast = fast, cdfname = cdfname,
                    read.verbose = read.verbose)
  }
  if(speed == "memory"){
    pms <- mem.bkg(filenames = filenames, pm.affinities = pm.affinities,
                    mm.affinities = mm.affinities, index.affinities = index.affinities,
                    type = type, minimum = minimum, optical.correct = optical.correct,
                    verbose = verbose, k = k, rho = rho, correction = correction,
                   stretch = stretch, fast = fast, cdfname = cdfname,
                   read.verbose = read.verbose)
  }

 
  tmp <- new("AffyBatch",
             cdfName=cdfName,
             annotation=cleancdfname(cdfName, addcdf=FALSE))
  
  ngenes <- length(featureNames(tmp))

  pNList <- probeNames(tmp)
  pNList <- split(0:(length(pNList) -1), pNList)


  
  exprs <- .Call("rma_c_complete",pms, pNList, ngenes, normalize, FALSE, bgversion, verbose, PACKAGE="affy")

  colnames(exprs) <- samplenames
  se.exprs <- array(NA, dim(exprs))
  
  annotation <- annotation(tmp)
  
  new("ExpressionSet",
      phenoData = phenoData,
      annotation = annotation,
      experimentData = description,
      exprs = exprs, se.exprs = se.exprs)
}

## A function to do background correction fast, but taking more RAM

fast.bkg <- function(filenames, pm.affinities, mm.affinities,
                     index.affinities, type, minimum, optical.correct,
                     verbose, k, rho, correction, stretch, fast, cdfname, read.verbose){
  
  pms <- read.probematrix(filenames=filenames, which="pm", cdfname = cdfname,
                          verbose = read.verbose)$pm
  mms <- read.probematrix(filenames=filenames, which="mm", cdfname = cdfname)$mm

  if(optical.correct){
     if(verbose) cat("Adjusting for optical effect.")
     for (i in 1:ncol(pms)){
       if(verbose) cat(".")
       tmp <- min(c(pms[,i], mms[,i]), na.rm=TRUE)
       pms[,i] <- pms[,i]- tmp + minimum
       mms[,i] <- mms[,i]- tmp + minimum
    }
     if(verbose) cat("Done.\n")
  }
  if(type=="fullmodel" | type=="affinities"){
    set.seed(1)
    Subset <- sample(1:length(pms[index.affinities,]),25000)
    y <- log2(pms)[index.affinities,][Subset]
    Subset <- (Subset-1)%%nrow(pms[index.affinities,])+1
    x <- pm.affinities[Subset]
    fit1 <- lm(y~x)
  }

  if(verbose) cat("Adjusting for non-specific binding")
  for(i in 1:ncol(pms)){
    if(verbose) cat(".")

          
    if(type=="fullmodel"){
      pms[,i] <- bg.adjust.fullmodel(pms[,i],mms[,i],ncs=NULL,
                                     pm.affinities,mm.affinities,anc=NULL,
                                     index.affinities,k=k,
                                     rho=rho,fast=fast)
      pms[,i] <- GSB.adj(pms[,i], index.affinities, pm.affinities, fit1$coef, k)
   #   pms[index.affinities,i] <- 2^(log2(pms[index.affinities,i])-
    #                                fit1$coef[2]*pm.affinities+mean(fit1$coef[2]*pm.affinities))
    }
    if(type=="affinities"){
      pms[,i] <- bg.adjust.affinities(pms[,i],ncs=mms[,i],pm.affinities,mm.affinities,
                                      index.affinities, k=k,
                                      fast=fast)
      pms[,i] <- GSB.adj(pms[,i], index.affinities, pm.affinities, fit1$coef, k)
      
      ## pms[index.affinities,i] <- 2^(log2(pms[index.affinities,i])-
##                                     fit1$coef[2]*pm.affinities + 
##                                    mean(fit1$coef[2]*pm.affinities))
    }
    if(type=="mm") pms[,i] <- bg.adjust.mm(pms[,i],correction*mms[,i],k=k,fast=fast)
    if(type=="constant"){
      pms[,i] <- bg.adjust.constant(pms[,i],k=k,Q=correction*mean(pms[,i]<mms[,i]),fast=fast)
    }
    if(stretch!=1){
      mu <- mean(log(pms[,i]))
      pms[,i] <- exp(mu + stretch*(log(pms[,i])-mu))
    }
  }
  if(verbose) cat("Done.\n")
  return(pms)
}

## A function to do background correction using less RAM

mem.bkg <- function(filenames, pm.affinities, mm.affinities,
                     index.affinities, type, minimum, optical.correct,
                     verbose, k, rho, correction, stretch, fast, cdfname, read.verbose){
  
  pms <- read.probematrix(filenames=filenames, which="pm", cdfname = cdfname,
                          verbose = read.verbose)$pm

  ## tmps used to carry optical correct value to bkg correction loop
   if(optical.correct){
     if(verbose) cat("Adjusting for optical effect.")
     tmps <- NULL
     for (i in 1:ncol(pms)){
       if(verbose) cat(".")
       mm <- read.probematrix(filenames=filenames[i], which="mm",
                              cdfname = cdfname)$mm[,1]
       tmp <-  min(c(pms[,i], mm), na.rm=TRUE)
       pms[,i] <- pms[,i]- tmp + minimum
       tmps <- c(tmps, tmp)
    }
   }
  if(verbose) cat("Done.\n")

  
  if(type=="fullmodel" | type=="affinities"){
    set.seed(1)
    Subset <- sample(1:length(pms[index.affinities,]),25000)
    y <- log2(pms)[index.affinities,][Subset]
    Subset <- (Subset-1)%%nrow(pms[index.affinities,])+1
    x <- pm.affinities[Subset]
    fit1 <- lm(y~x)
  }

  if(verbose) cat("Adjusting for non-specific binding")
  for(i in 1:ncol(pms)){
    if(verbose) cat(".")

    mm <- read.probematrix(filenames=filenames[i], which="mm", cdfname = cdfname)$mm[,1]
    
    if(optical.correct)
      mm <- mm - tmps[i] + minimum
          
    if(type=="fullmodel"){
      pms[,i] <- bg.adjust.fullmodel(pms[,i],mm,ncs=NULL,
                                     pm.affinities,mm.affinities,anc=NULL,
                                     index.affinities,k=k,
                                     rho=rho,fast=fast)
      pms[index.affinities,i] <- 2^(log2(pms[index.affinities,i])-
                                    fit1$coef[2]*pm.affinities+mean(fit1$coef[2]*pm.affinities))
    }
    if(type=="affinities"){
      pms[,i] <- bg.adjust.affinities(pms[,i],ncs=mm,pm.affinities,mm.affinities,
                                      index.affinities, k=k,
                                      fast=fast)
      pms[index.affinities,i] <- 2^(log2(pms[index.affinities,i])-
                                    fit1$coef[2]*pm.affinities + 
                                    mean(fit1$coef[2]*pm.affinities))
    }
    if(type=="mm") pms[,i] <- bg.adjust.mm(pms[,i],correction*mm,k=k,fast=fast)
    if(type=="constant"){
      pms[,i] <- bg.adjust.constant(pms[,i],k=k,Q=correction*mean(pms[,i]<mm),fast=fast)
    }
    if(stretch!=1){
      mu <- mean(log(pms[,i]))
      pms[,i] <- exp(mu + stretch*(log(pms[,i])-mu))
    }
  }
  if(verbose) cat("Done.\n")
  return(pms)
}


  

gcrma <- function(object,affinity.info=NULL,
                  affinity.source=c("reference","local"),
                  NCprobe=NULL,
                  type=c("fullmodel","affinities","mm","constant"),
                  k=6*fast+0.5*(1-fast),stretch=1.15*fast+1*(1-fast),correction=1,
                  GSB.adjust=TRUE,rho=0.7,
                  optical.correct=TRUE,verbose=TRUE,fast=TRUE,
                  subset=NULL,normalize=TRUE,...){
  object <- bg.adjust.gcrma(object,
                            affinity.info=affinity.info,
                            affinity.source=affinity.source, NCprobe=NCprobe,
                            type=type,k=k,stretch=stretch,correction=correction,
                            GSB.adjust=GSB.adjust,rho=rho,
                            optical.correct=optical.correct,verbose=verbose,fast=fast)
  return(rma(object,subset=subset,background=FALSE,normalize=normalize,verbose=verbose))
}



bg.adjust.gcrma <- function(object,affinity.info=NULL,
                            affinity.source=c("reference","local"),
                            NCprobe=NULL,
                            type=c("fullmodel","affinities","mm","constant"),
                            k=6*fast+0.5*(1-fast),stretch=1.15*fast+1*(1-fast),correction=1,
                            GSB.adjust=TRUE,rho=0.7,
                            optical.correct=TRUE,verbose=TRUE,fast=TRUE){
  ##OPTICAL BG
  if(optical.correct)
    object <- bg.adjust.optical(object,verbose=verbose)
  type <- match.arg(type)
  pmonly <- (type=="affinities"|type=="constant")
  needaff <- (type=="fullmodel"|type=="affinities")
  ##OBTAIN AFFINITY.INFO

  Source=match.arg(affinity.source)
  if( needaff & is.null(affinity.info) & Source=="reference" )  ##use the GCRMA internal NSB data 
    affinity.info <- compute.affinities(cdfName(object),verbose=verbose)
  if( needaff & is.null(affinity.info) & Source=="local" )  ##use experimental data
    affinity.info <- compute.affinities.local(object,Array=NULL,NCprobe=NCprobe,
                                              verbose=verbose,optical.correct=FALSE)

  ##clean NCprobe
  if(!is.null(NCprobe)){
    if(length(affinity.info)==1){
      anc=as.matrix(intensity(affinity.info)[NCprobe,])
      index2=which(!is.na(anc))
      anc=anc[index2,]
      ncs=as.matrix(intensity(object)[NCprobe[index2],])}
    else {
      anc=intensity(affinity.info)[NCprobe,]
      index2=which(!is.na(anc[,1]))
      anc=anc[index2,]
      ncs=as.matrix(intensity(object)[NCprobe[index2],])
    }
  }
  else anc=ncs=NULL
  
  pmIndex=unlist(indexProbes(object,"pm"))
  mmIndex=unlist(indexProbes(object,"mm"))
  
  if(!is.null(affinity.info)){
    if(length(affinity.info)==1){ #one array affinity.info, NCprobe=MM
      pm(object) <- gcrma.engine(pms=pm(object),
                                 mms=mm(object),ncs,
                                 pm.affinities=pm(affinity.info),
                                 mm.affinities=mm(affinity.info),anc,
                                 type=type,k=k,
                                 stretch=stretch,
                                 correction=correction,GSB.adjust=GSB.adjust,rho=rho,
                                 verbose=verbose,fast=fast)
    }
    else if (length(affinity.info)==length(object)){#multi-array affinity.info
      pm(object) <- gcrma.engine2(object,pmIndex,mmIndex,
                                  NCprobe=NCprobe,
                                  affinity.info,
                                  type=type,k=k,
                                  stretch=stretch,
                                  correction=correction,GSB.adjust=GSB.adjust,rho=rho,
                                  verbose=verbose,fast=fast)}
  }
  else{#local affinity.info, not provided.
    
  }
  return(object)
}

bg.parameters.ns=function (x, affinities,affinities2=NULL,affinities3=NULL,span=.2)
{   
#  set.seed(1989)
  O=order(affinities)
  x1<-runmed(x[O],21)
  a1 <- affinities[O]
  window1 <- cut(a1, c(-Inf,seq(quantile(a1,.001),quantile(a1,.999),len=26),Inf))
  sample1= tapply(1:length(a1),window1,function(x)
    floor(quantile(x,seq(0,1,len=min(length(x),200)))))
  sample1=unlist(sample1[2:26])
  suppressWarnings(lo1<-loess(log(x1)~a1,subset=sample1,degree=1,family="symmetric",span=span))
  bg.mu=predict(lo1,affinities)
  bg.mu[affinities>max(a1[sample1])]=max(lo1$fitted)
  bg.mu[affinities<min(a1[sample1])]=min(lo1$fitted)
  res=log(x)-bg.mu;res=res[res<0];res=c(res,-res)
  bg.sigma=mad(res)

  if (is.null(affinities2)) bg.mu2=NULL
  else {bg.mu2=predict(lo1,affinities2)
        bg.mu2[affinities2>max(a1[sample1])]=max(lo1$fitted)
        bg.mu2[affinities2<min(a1[sample1])]=min(lo1$fitted)
      }
  
  if (is.null(affinities3))        return(list(bg.mu=bg.mu,bg.mu2=bg.mu2,bg.sigma=bg.sigma))
  
  else {bg.mu3=predict(lo1,affinities3)
        bg.mu3[affinities3>max(a1[sample1])]=max(lo1$fitted)
        bg.mu3[affinities3<min(a1[sample1])]=min(lo1$fitted)
        return(list(bg.mu=bg.mu,bg.mu2=bg.mu2,bg.mu3=bg.mu3,bg.sigma=bg.sigma))
      }
}


# sg.parameters=function(pms,apm,Source){
#   apm=as.matrix(apm)
#   index.affinities <- which(!is.na(apm[,1]))
#   apm=apm[index.affinities,]
#   pms=pms[index.affinities,]
#   set.seed(1)
#   if(Source=="reference"){ #one array
#     Subset <- sample(1:length(pms),25000)
#     y <- log2(pms)[Subset]
#     Subset <- (Subset-1)%%nrow(pms)+1
#     x <- apm[Subset]}
#   else {#multiple arrays
#     set.seed(1)
#     Subset <- sample(1:length(pms),25000)
#     y <- log2(pms)[Subset]
#     x <- apm[Subset]}
#   fit1 <- lm(y~x)
#   return(fit1$coef)
# }
  
bg.parameters.ns2 <- function (x, affinities,affinities2=NULL,affinities3=NULL,span=.2)
{   
#  set.seed(1989)
  O=order(affinities)
  x1<-runmed(x[O],21)
  a1 <- affinities[O]
  window1 <- cut(a1, c(-Inf,seq(quantile(a1,.001),quantile(a1,.999),len=26),Inf))
  sample1= tapply(1:length(a1),window1,function(x)
    floor(quantile(x,seq(0,1,len=min(length(x),200)))))
  sample1=unlist(sample1[2:26])
  suppressWarnings(lo1<-loess(log(x1)~a1,subset=sample1,degree=1,family="symmetric",span=.2))
  bg.mu=predict(lo1,affinities)
  bg.mu[affinities>max(a1[sample1])]=max(lo1$fitted)
  bg.mu[affinities<min(a1[sample1])]=min(lo1$fitted)
  res=log(x)-bg.mu;res=res[res<0];res=c(res,-res)
  bg.sigma=mad(res)

  if (is.null(affinities2)) bg.mu2=NULL
  else {bg.mu2=predict(lo1,affinities2)
        bg.mu2[affinities2>max(a1[sample1])]=max(lo1$fitted)
        bg.mu2[affinities2<min(a1[sample1])]=min(lo1$fitted)
      }
  
  if (is.null(affinities3))        return(list(bg.mu=bg.mu,bg.mu2=bg.mu2,bg.sigma=bg.sigma))
  
  else {bg.mu3=predict(lo1,affinities3)
        bg.mu3[affinities3>max(a1[sample1])]=max(lo1$fitted)
        bg.mu3[affinities3<min(a1[sample1])]=min(lo1$fitted)
        return(list(bg.mu=bg.mu,bg.mu2=bg.mu2,bg.mu3=bg.mu3,bg.sigma=bg.sigma))
      }
}

  

##multi-array affinity.info
gcrma.engine2 <- function(object,pmIndex=NULL,mmIndex=NULL,
                          NCprobe=NULL,
                          affinity.info,
                          type=c("fullmodel","affinities","mm","constant"),
                          k=6*fast+0.5*(1-fast),
                          stretch=1.15*fast+1*(1-fast),correction=1,GSB.adjust=TRUE,rho=0.7,
                          verbose=TRUE,fast=TRUE){

  if(is.null(pmIndex)) pmIndex=unlist(indexProbes(object, "pm"))
  if(is.null(mmIndex)) mmIndex=unlist(indexProbes(object, "mm"))
  pm.affinities=intensity(affinity.info)[pmIndex,]
  mm.affinities=intensity(affinity.info)[mmIndex,]
  pms=intensity(object)[pmIndex,]
  mms=intensity(object)[mmIndex,]
  if(!is.null(pm.affinities)){
    index.affinities <- which(!is.na(pm.affinities[,1]))
    pm.affinities <- pm.affinities[index.affinities,]
    if(!is.null(mm.affinities)){
      mm.affinities <- mm.affinities[index.affinities,]
    }
  }
  
###get GSB.adjust parameters #multiple arrays
  if(GSB.adjust &(type=="fullmodel" | type=="affinities")){
    set.seed(1)
    Subset <- sample(1:length(pms[index.affinities,]),25000)
    y <- log2(pms)[index.affinities,][Subset]
    x <- pm.affinities[Subset]
    fit1 <- lm(y~x)$coef
  }

  if(verbose) cat("Adjusting for non-specific binding")
  for(i in 1:ncol(pms)){
    if(verbose) cat(".")
    ##fullmodel
    if(type=="fullmodel") {
      if(is.null(NCprobe)) 
        pms[,i] <- bg.adjust.fullmodel(pms[,i],mms[,i],ncs=NULL,
                                       apm=pm.affinities[,i],amm=mm.affinities[,i],
                                       anc=NULL,
                                       index.affinities,k=k,
                                       rho=rho,fast=fast)
    
      else
        pms[,i] <- bg.adjust.fullmodel(pms[,i],mms[,i],ncs=intensity(object)[NCprobe,i],
                                       apm=pm.affinities[,i],amm=mm.affinities[,i],
                                       anc=intensity(affinity.info)[NCprobe,i],
                                       index.affinities,k=k,
                                       rho=rho,fast=fast)
    
      if(GSB.adjust)
        pms[,i] <- GSB.adj(Yin=pms[,i],subset=index.affinities,aff=pm.affinities[,i],fit1=fit1,k=k)

    }  
    ##affinities
    if(type=="affinities"){
      if(is.null(NCprobe))
        pms[,i] <-  bg.adjust.affinities(pms[,i],ncs=mms[,i],
                                         apm=pm.affinities[,i],anc=mm.affinities[,i],
                                         index.affinities,k=k,fast=fast)
      else
        pms[,i] <- bg.adjust.affinities(pms[,i],ncs=intensity(object)[NCprobe,i],
                                        apm=pm.affinities[,i],anc=intensity(affinity.info)[,i],
                                        index.affinities,k=k,fast=fast,nomm=TRUE)
      if(GSB.adjust)
        pms[,i] <- GSB.adj(Yin=pms[,i],subset=index.affinities,aff=pm.affinities,fit1=fit1,k=k)
      }
    if(type=="mm") pms[,i] <- bg.adjust.mm(pms[,i],correction*mms[,i],k=k,fast=fast)
    if(type=="constant"){
      pms[,i] <- bg.adjust.constant(pms[,i],k=k,Q=correction*mean(pms[,i]<mms[,i]),fast=fast)
    }
 
    if(stretch!=1){
      mu <- mean(log(pms[,i]))
      pms[,i] <- exp(mu + stretch*(log(pms[,i])-mu))
    }
  }

  if(verbose) cat("Done.\n")
  return(pms)
}

\name{compute.affinities}
\alias{compute.affinities}
\alias{compute.affinities2}
\alias{check.probes}
\alias{base.profiles}
\alias{base.profiles.mm}
\alias{base.profiles.nc}
\alias{compute.affinity.coef}
\alias{compute.affinities.local}
\alias{plotBaseProfiles}

\title{Probe Affinity computation}
\description{
An internal function to calculate probe affinities from their
sequences.}
\usage{
compute.affinities(cdfname,verbose=TRUE)
compute.affinities2(cdfname,verbose=TRUE)
check.probes(probepackage,cdfname)
}
\arguments{
  \item{cdfname}{ Object of class \code{character} representing the name of
          \code{CDF} file associated with the arrays in the \code{AffyBatch}.}
\item{probepackage}{\code{character} representing the name of the
  package with the probe sequence information.}
 \item{verbose}{Logical value. If \code{TRUE} messages about the progress of
    the function is printed.} 

}
\details{
The affinity of a probe is described as the sum of position-dependent base affinities.
Each base at each position contributes to the total affinity of a probe
in an additive fashion. For a given type of base, the positional effect
is modeled as a spline function with 5 degrees of freedom.

Use \code{compute.affinities2} if there are no MM probes.

\code{check.probes} makes sure things are matching as they should.
}

\value{\code{compute.affinities} returns an \code{AffyBatch} with the
  affinities for PM probes in the pm locations and the affinities for MM
  probes in the mm locations. NA  will be added for probes with no
  sequence information.
}
\references{
Hekstra, D., Taussig, A. R., Magnasco, M., and Naef, F. (2003) Absolute mRNA concentrations from sequence-specific calibration of oligonucleotide array. Nucleic Acids Research, 31. 1962-1968.
}
\seealso{\code{\link{gcrma},\link{affinity.spline.coefs}}}
\author{Rafeal Irizarry}
\keyword{manip}% at least one, from doc/KEYWORDS

\name{affinity.spline.coefs}
\docType{data}
\alias{affinity.spline.coefs}
\title{Spline coefficients for estimation of affinity from probe sequence}
\description{Spline coefficients for estimation of affinity from probe sequence}
\usage{data(affinity.spline.coefs)}
%\format{}
%\source{}
%\url{}
\seealso{\code{\link{compute.affinities}}}
\keyword{datasets}

\name{bg.adjust.affinities}
\alias{bg.adjust.affinities}
\alias{bg.adjust.mm}
\alias{bg.adjust.fullmodel}
\alias{bg.adjust.constant}
\alias{bg.adjust.optical}
\title{Background adjustment with sequence information (internal function)}
\description{
An internal function to be used by \code{\link{gcrma}}.
}
\usage{
bg.adjust.fullmodel(pms,mms,ncs=NULL,apm,amm,anc=NULL,index.affinities,k=6
* fast + 0.25 * (1 - fast),rho=.7,fast=FALSE)
bg.adjust.affinities(pms,ncs,apm,anc,index.affinities,k=6
* fast + 0.25 * (1 - fast),fast=FALSE,nomm=FALSE)
}
\arguments{
  \item{pms}{PM intensities after optical background correction, before
    non-specific-binding correction.}
  \item{mms}{MM intensities after optical background correction, before
    non-specific-binding correction.}
  \item{ncs}{Negative control probe intensities after optical background correction, before
    non-specific-binding correction. If \code{ncs=NULL}, the MM probes
    are considered the negative control probes.}
  \item{index.affinities}{The index of pms with known sequences. (For some types of
    arrays the sequences of a small subset of probes are not provided by
    Affymetrix.)}
  \item{apm}{Probe affinities for PM probes with known sequences.}
  \item{amm}{Probe affinities for MM probes with known sequences.}
  \item{anc}{Probe affinities for Negative control probes with known
    sequences. This is ignored when \code{ncs=NULL}.}
  \item{rho}{correlation coefficient of log background intensity in a pair of pm/mm probes. Default=.7}
  \item{k}{A tuning parameter. See details.}
%  \item{Q}{A number between 0 and 1 that determines what quantile to use
%  as an estimate of the mean background noise.}
  \item{fast}{Logical value. If \code{TRUE} a faster add-hoc algorithm is used.}
  \item{nomm}{Logical value indicating if MM intensities are available
    and will to be used to estimate background.}
}
\details{Assumes PM=background1+signal,mm=background2,
  (log(background1),log(background2))' 
  follow bivariate normal distribution, signal distribution follows power
  law. 
  \code{bg.parameters.gcrma} and \code{sg.parameters.gcrma} 
  provide adhoc estimates of the parameters.

  the original gcrma uses an empirical Bayes estimate. this requires a
  complicated numerical integration. An add-hoc method tries to imitate
  the empirical Bayes estimate with a PM-B but values of PM-B<\code{k}
  going to \code{k}. This can be thought as a shrunken MVUE. For more
  details see Wu et al. (2003). 
}
\value{a vector of same length as x.}
\seealso{\code{\link{gcrma}}}
\author{Rafeal Irizarry, Zhijin(Jean) Wu}
\keyword{manip}% at least one, from doc/KEYWORDS

\name{justGCRMA}
\alias{justGCRMA}
\alias{just.gcrma}
\title{Compute GCRMA Directly from CEL Files}
\description{
  This function converts CEL files into an \code{ExpressionSet}
  using the robust multi-array average (RMA) expression measure with help of probe sequences.
}
\usage{
            just.gcrma(\dots, filenames=character(0),
                       phenoData=new("AnnotatedDataFrame"),
                       description=NULL,
                       notes="", compress=getOption("BioC")$affy$compress.cel,
                       normalize=TRUE, bgversion=2, affinity.info=NULL,
                       type=c("fullmodel","affinities","mm","constant"),
                       k=6*fast+0.5*(1-fast), stretch=1.15*fast+1*(1-fast),
                       correction=1, rho=0.7, optical.correct=TRUE,
                       verbose=TRUE, fast=TRUE, minimum=1, optimize.by =
                       c("speed","memory"), 
                       cdfname = NULL, read.verbose = FALSE)

            justGCRMA(\dots, filenames=character(0),
                     widget=getOption("BioC")$affy$use.widgets,
                     compress=getOption("BioC")$affy$compress.cel,
                     celfile.path=getwd(),
                     sampleNames=NULL,
                     phenoData=NULL,
                     description=NULL,
                     notes="",
                     normalize=TRUE, 
                     bgversion=2, affinity.info=NULL,
                     type=c("fullmodel","affinities","mm","constant"),
                     k=6*fast+0.5*(1-fast), stretch=1.15*fast+1*(1-fast),
                     correction=1, rho=0.7, optical.correct=TRUE,
                     verbose=TRUE, fast=TRUE, minimum=1,
                     optimize.by = c("speed","memory"),
                     cdfname = NULL, read.verbose = FALSE)
}
\arguments{
  \item{\dots}{file names separated by comma.}
  \item{filenames}{file names in a character vector.}
  \item{widget}{a logical specifying if widgets should be used.}
  \item{compress}{are the CEL files compressed?}
  \item{phenoData}{a \code{\link[Biobase:class.AnnotatedDataFrame]{AnnotatedDataFrame}} object.}
  \item{description}{a \code{\link[Biobase:class.MIAME]{MIAME}} object.}
  \item{notes}{notes.}
  \item{affinity.info}{\code{NULL} or a list of three components:
    apm,amm and index, for PM probe affinities, MM probe affinities,
    the index of probes with known sequence, respectively.}
  \item{type}{"fullmodel" for sequence and MM model. "affinities" for
    sequence information only. "mm" for using MM without sequence
    information.}
  \item{k}{A tuning factor.}
  \item{rho}{correlation coefficient of log background intensity in a pair of pm/mm probes.
    Default=.7.}
  \item{stretch}{.}
  \item{correction}{.}
  \item{normalize}{Logical value. If \code{TRUE}, then normalize data using
    quantile normalization.}
  \item{optical.correct}{Logical value. If \code{TRUE}, then optical
    background correction is performed.}
  \item{verbose}{Logical value. If \code{TRUE}, then messages about the progress of
    the function is printed.} 
  \item{fast}{Logical value. If \code{TRUE}, then a faster add-hoc algorithm is used.}
  \item{optimize.by}{"speed" will use a faster algorithm but more RAM, and "memory"
    will be slower, but require less RAM.}
  \item{bgversion}{integer value indicating which RMA background to use 1: use
          background similar to pure R rma background given in affy
          version 1.0 - 1.0.2 2: use background similar to pure R rma
          background given in affy version 1.1 and above.} 
  \item{minimum}{.}
  \item{celfile.path}{a character denoting the path 'ReadAffy' should look for
          cel files.}
  \item{sampleNames}{a character vector of sample names to be used in the
    'AffyBatch'.}
  \item{cdfname}{Used to specify the name of an alternative cdf package. If set to
     \code{NULL}, the usual cdf package based on Affymetrix' mappings
     will be used. Note that the name should not include the 'cdf' on
     the end, and that the corresponding probe package is also required
     to be installed. If either package is missing an error will
     result.}
   \item{read.verbose}{Logical value. If \code{TRUE}, then messages will be
     printed as each celfile is read in.}
}
\details{
  This method should require much less RAM than the conventional
  method of first creating an \code{AffyBatch} and then running
  \code{\link[gcrma]{gcrma}}.

  This is a simpler version than \code{gcrma}, so some of the arguments
  available in \code{gcrma} are not available here. For example, it is
  not possible to use the MM probes to estimate background. Instead, the
  internal NSB estimates are used (which is also the default for \code{gcrma}).
  
  Note that this expression measure is given to you in log base 2
  scale. This differs from most of the other expression measure
  methods.

  The tuning factor \code{k} will have different meanings if one uses
  the fast (add-hoc) algorithm or the empirical Bayes approach. See Wu
  et al. (2003)

  \code{fast.bkg} and \code{mem.bkg} are two internal functions.
}
\value{
 An \code{\link[Biobase:class.ExpressionSet]{ExpressionSet}} object.
}
\author{James W. MacDonald}
\keyword{manip}

\name{bg.parameters.ns}
\alias{bg.parameters.ns}
\alias{left.sigma}
\alias{PAV}
\alias{average.for.PAV}
\title{Estimation of non-specific Binding Background Parameters}
\description{
An internal function to be used by \code{\link{gcrma}}
}
\usage{
bg.parameters.ns(x,affinities,affinities2=NULL,affinities3=NULL,span=.2)
}
\arguments{
  \item{x}{PM or MM intensities after optical background correction, before non-specific-binding correction.}
  \item{affinities}{Probe affinities for probes with known
    sequences.Used to estimate the function between non-specific binding
  and affinities.}
\item{affinities2}{
   Probe affinities for the probes whoes expected non-specific binding
   intensity is to be predicted.
}  
 \item{affinities3}{
   Probe affinities for another extra group of probes whoes expected non-specific binding
   intensity is to be predicted.
 }
 \item{span}{The span parameter passed to loess function}
}
%\details{}
\value{a vector of same length as x.}
\seealso{\code{\link{gcrma}}}
\author{Rafeal Irizarry, Zhijin (Jean) Wu}
\keyword{manip}

\name{bg.adjust.gcrma}
\alias{bg.adjust.gcrma}
\title{GCRMA background adjust (internal function)}
\description{
  This function performs background adjustment (optical noise and
  non-specific binding on an \code{AffyBatch}
  project and returns an \code{AffyBatch} object in which the PM
  intensities are adjusted.
}
\usage{
bg.adjust.gcrma(object,affinity.info=NULL,
      affinity.source=c("reference","local"),
      NCprobe=NULL,
      type=c("fullmodel","affinities","mm","constant"),
      k=6*fast+0.5*(1-fast),stretch=1.15*fast+1*(1-fast),correction=1,
      GSB.adjust=TRUE,
      rho=.7,optical.correct=TRUE,verbose=TRUE,fast=TRUE)
}
\arguments{
  \item{object}{an \code{\link[affy:AffyBatch-class]{AffyBatch}}}
  \item{affinity.info}{\code{NULL} or an \code{AffyBatch} containing the
    affinities in the \code{exprs} slot. This object can be created
    using the function \code{\link{compute.affinities}}.}
  \item{affinity.source}{\code{reference}: use the package internal
    Non-specific binding data or \code{local}: use the experimental
    data in \code{object}. If \code{local} is chosen, either MM probes or a user-defined
    list of probes (see \code{NCprobes}) are used to estimate affinities.}
  \item{NCprobe}{Index of negative control probes. When set as
    \code{NULL},the MM probes will be used. These probes
    are used to estimate parameters of non-specific binding on each
    array. These will be also used to estimate probe affinity profiles when
    affinity.info is not provided.}
  \item{type}{"fullmodel" for sequence and MM model. "affinities" for
    sequence information only. "mm" for using MM without sequence
    information.}
  \item{k}{A tuning factor.}
  \item{stretch}{.}
  \item{correction}{.}
  \item{GSB.adjust}{Logical value. If \code{TRUE}, probe effects in specific binding will
    be adjusted.}
  \item{rho}{correlation coefficient of log background intensity in a pair of pm/mm probes. Default=.7}
  \item{optical.correct}{Logical value. If \code{TRUE}, optical
    background correction is performed.}
  \item{verbose}{Logical value. If \code{TRUE} messages about the progress of
    the function is printed.} 
  \item{fast}{Logical value. If \code{TRUE} a faster ad hoc algorithm is
    used.}
}
\details{
  The returned value is an \code{AffyBatch} object, in which the PM probe intensities
  have been background adjusted. The rest is left the same as the
  starting \code{AffyBatch} object.
  
  The tunning factor \code{k} will have different meainngs if one uses
  the fast (ad hoc) algorithm or the empirical bayes approach. See Wu
  et al. (2003)
    
}
\value{
 An \code{AffyBatch}.
}
\author{Rafeal Irizarry}
\examples{
 if(require(affydata) & require(hgu95av2probe) & require(hgu95av2cdf)){
          data(Dilution)
          ai <- compute.affinities(cdfName(Dilution))
          Dil.adj<-bg.adjust.gcrma(Dilution,affinity.info=ai,type="affinities")
     }
}
\keyword{manip}

\name{getCDF}
\alias{getCDF}
\alias{getProbePackage}
%- Also NEED an '\alias' for EACH other topic documented here.
\title{ Functions for Automatic Download of Packages }
\description{
  These are internal functions that are called by \code{justGCRMA} and
  \code{GCRMA} in order to automatically download and install cdf
  environments and probe packages.
}
\usage{
getCDF(cdfname, lib = .libPaths()[1], verbose = TRUE)
getProbePackage(probepackage, lib = .libPaths()[1], verbose = TRUE)
}
%- maybe also 'usage' for other objects documented here.
\arguments{
  \item{cdfname}{ Name of the cdfenv to install. }
  \item{probepackage}{ Name of the probe package to install. }
  \item{lib}{ Directory of the R library where packages will be installed. }
  \item{verbose}{ Output informative comments? Defaults to \code{TRUE}}
}

\value{
  Nothing is returned. These functions are called simply to install
  environments. 
}
\author{ James W. MacDonald, based on \code{getCDFinfo}, written by Jeff
Gentry.}

\seealso{ \code{getCDFinfo} 
  %\code{\link[affy]{getCDFinfo}} there is no such function in affy
  %called getCDFinfo 
}

\keyword{ internal }

\name{gcrma.engine2}
\alias{gcrma.engine2}
\title{GCRMA background adjust engine(internal function)}
\description{
  This function adjust for non-specific binding when each array has its
  own probe affinity information. It takes an AffyBatch object of probe
  intensities and an AffyBatch of probe  affinity,
  returns one matrix of non-specific binding corrected PM probe intensities.
}
\usage{
gcrma.engine2(object,pmIndex=NULL,mmIndex=NULL,
              NCprobe=NULL,affinity.info,
              type=c("fullmodel","affinities","mm","constant"),
              k=6*fast+0.5*(1-fast),
              stretch=1.15*fast+1*(1-fast),correction=1,GSB.adjust=TRUE,rho=0.7,
              verbose=TRUE,fast=TRUE)
}
\arguments{
  \item{object}{an \code{\link[affy:AffyBatch-class]{AffyBatch}}. Note: this is an internal
    function. Optical noise should have been corrected for. }
  \item{pmIndex}{Index of PM probes.This will be computed within the
    function if left \code{NULL}}
  \item{mmIndex}{Index of MM probes.This will be computed within the
    function if left \code{NULL}}
  \item{NCprobe}{Index of negative control probes. When set as
    \code{NULL},the MM probes will be used. These probes
    are used to estimate parameters of non-specific binding on each
    array. These will be also used to estimate probe affinity profiles when
    affinity.info is not provided.}
  \item{affinity.info}{\code{NULL} or an \code{AffyBatch} containing the
    affinities in the \code{exprs} slot. This object can be created
    using the function \code{\link{compute.affinities}}.}
  \item{type}{"fullmodel" for sequence and MM model. "affinities" for
    sequence information only. "mm" for using MM without sequence
    information.}
  \item{k}{A tuning factor.}
  \item{stretch}{.}
  \item{correction}{.}
  \item{GSB.adjust}{Logical value. If \code{TRUE}, probe effects in specific binding will
    be adjusted.}
  \item{rho}{correlation coefficient of log background intensity in a pair of pm/mm probes. Default=.7}
  \item{verbose}{Logical value. If \code{TRUE} messages about the progress of
    the function is printed.} 
  \item{fast}{Logicalvalue. If \code{TRUE} a faster add-hoc algorithm is
    used.}
}
\details{
  Note that this expression measure is given to you in log base 2
  scale. This differs from most of the other expression measure
  methods.

  The tunning factor \code{k} will have different meainngs if one uses
  the fast (add-hoc) algorithm or the empirical bayes approach. See Wu
  et al. (2003)
     
}
\value{
 A matrix of PM intensties. 
}
\seealso{gcrma.engine}
\author{Rafeal Irizarry & Zhijin Wu}
\keyword{manip}

\name{fast.bkg}
\alias{fast.bkg}
\alias{mem.bkg}
%- Also NEED an '\alias' for EACH other topic documented here.
\title{Internal functions for justGCRMA}
\description{
  These are internal functions for justGCRMA that are called based on
  memory or speed constraints.
}
\usage{
fast.bkg(filenames, pm.affinities, mm.affinities, index.affinities,
type, minimum, optical.correct, verbose, k, rho, correction, stretch,
fast, cdfname, read.verbose)
mem.bkg(filenames, pm.affinities, mm.affinities, index.affinities, type,
minimum, optical.correct, verbose, k, rho, correction, stretch, fast,
cdfname, read.verbose)
}
%- maybe also 'usage' for other objects documented here.
\arguments{
  \item{filenames}{A list of cel files.}
  \item{pm.affinities}{Values passed from \code{compute.affinities}.}
  \item{mm.affinities}{Values passed from \code{compute.affinities}.}
  \item{index.affinities}{Values passed from \code{compute.affinities}.}
  \item{type}{"fullmodel" for sequence and MM model. "affinities" for
          sequence information only. "mm" for using MM without sequence
          information.}
  \item{minimum}{A minimum value to be used for \code{optical.correct}.}
  \item{optical.correct}{Logical value. If \code{TRUE}, optical
    background correction is performed.}
  \item{verbose}{Logical value. If \code{TRUE}, messages about the
    progress of the function are printed.}
  \item{k}{A tuning factor.}
  \item{rho}{correlation coefficient of log background intensity in a pair
    of pm/mm probes. Default=.7}
  \item{correction}{  }
  \item{stretch}{  }
  \item{fast}{Logical value. If \code{TRUE}, then a faster ad hoc
    algorithm is used.}
  \item{cdfname}{Used to specify the name of an alternative cdf package. If set to
    \code{NULL}, the usual cdf package based on Affymetrix' mappings
    will be used. Note that the name should not include the 'cdf' on
    the end, and that the corresponding probe package is also required
    to be installed. If either package is missing an error will result.}
  \item{read.verbose}{Logical value. If \code{TRUE}, a message is
    returned as each celfile is read in.}
}
\details{
  Note that this expression measure is given to you in log base 2
  scale. This differs from most of the other expression measure
  methods.

  The tuning factor 'k' will have different meanings if one uses
  the fast (add-hoc) algorithm or the empirical Bayes approach. See
  Wu et al. (2003)
}
\value{
  An \code{ExpressionSet}.
}
\author{James W. MacDonald <jmacdon@med.umich.edu>}
\seealso{\code{\link{gcrma}}}
\keyword{internal}% at least one, from doc/KEYWORDS

\name{gcrma.engine}
\alias{gcrma.engine}
\title{GCRMA background adjust engine(internal function)}
\description{
  This function adjust for non-specific binding when all arrays in the
  dataset share the same probe affinity information. It takes matrices of PM
  probe intensities, MM probe intensities, other negative control probe
  intensities(optional) and the associated probe affinities, and return
  one matrix of non-specific binding corrected PM probe intensities.
}
\usage{
gcrma.engine(pms,mms,ncs=NULL,
                         pm.affinities=NULL,mm.affinities=NULL,anc=NULL,
                         type=c("fullmodel","affinities","mm","constant"),
                         k=6*fast+0.5*(1-fast),
                         stretch=1.15*fast+1*(1-fast),correction=1,GSB.adjust=TRUE,rho=0.7,
                         verbose=TRUE,fast=FALSE)
}
\arguments{
  \item{pms}{The matrix of PM intensities}
  \item{mms}{The matrix of MM intensities}
  \item{ncs}{The matrix of negative control probe intensities. When left
as\code{NULL}, the MMs are considered the negative control probes.}
  \item{pm.affinities}{The vector of PM probe affinities. Note: This can be
shorter than the number of rows in \code{pms} when some probes do not
have sequence information provided.}
 \item{mm.affinities}{The vector of MM probe affinities.}
 \item{anc}{The vector of Negative Control probe affinities. This is
ignored if MMs are used as negative controls (\code{ncs=NULL})}
  \item{type}{"fullmodel" for sequence and MM model. "affinities" for
    sequence information only. "mm" for using MM without sequence
    information.}
  \item{k}{A tuning factor.}
  \item{stretch}{.}
  \item{correction}{.}
  \item{GSB.adjust}{Logical value. If \code{TRUE}, probe effects in specific binding will
    be adjusted.}
  \item{rho}{correlation coefficient of log background intensity in a pair of pm/mm probes. Default=.7}
  \item{verbose}{Logical value. If \code{TRUE} messages about the progress of
    the function is printed.} 
  \item{fast}{Logicalvalue. If \code{TRUE} a faster add-hoc algorithm is
    used.}
}
\details{
  Note that this expression measure is given to you in log base 2
  scale. This differs from most of the other expression measure
  methods.

  The tunning factor \code{k} will have different meainngs if one uses
  the fast (add-hoc) algorithm or the empirical bayes approach. See Wu
  et al. (2003)
     
}
\value{
 A matrix of PM intensties. 
}
\seealso{gcrma.engine2}
\author{Rafeal Irizarry & Zhijin Wu}
\keyword{manip}

\name{gcrma}
\alias{gcrma}
\alias{gcrma.bg.transformation.fast}
\alias{gcrma.bg.transformation}
\alias{GSB.adj}
\title{Robust Multi-Array expression measure using sequence information}
\description{
  This function converts an \code{AffyBatch} into an \code{ExpressionSet}
  using the robust multi-array average (RMA) expression measure with help of probe sequence.
}
\usage{
gcrma(object,affinity.info=NULL,
      affinity.source=c("reference","local"),NCprobe=NULL,
      type=c("fullmodel","affinities","mm","constant"),
      k=6*fast+0.5*(1-fast),stretch=1.15*fast+1*(1-fast),correction=1,
      GSB.adjust=TRUE,
      rho=.7,optical.correct=TRUE,verbose=TRUE,fast=TRUE,
      subset=NULL,normalize=TRUE,\dots)
}
\arguments{
  \item{object}{an \code{\link[affy:AffyBatch-class]{AffyBatch}}}
  \item{affinity.info}{\code{NULL} or an \code{AffyBatch} containing the
    affinities in the \code{exprs} slot. This object can be created
    using the function \code{\link{compute.affinities}}.}
  \item{affinity.source}{\code{reference}: use the package internal
    Non-specific binding data or \code{local}: use the experimental
    data in \code{object}. If \code{local} is chosen, either MM probes or a user-defined
    list of probes (see \code{NCprobes}) are used to estimate
    affinities.}
  \item{NCprobe}{Index of negative control probes. When set as
    \code{NULL},the MM probes will be used. These probes
    are used to estimate parameters of non-specific binding on each
    array. These will be also used to estimate probe affinity profiles when
    affinity.info is not provided.}
  \item{type}{"fullmodel" for sequence and MM model. "affinities" for
    sequence information only. "mm" for using MM without sequence
    information.}
  \item{k}{A tuning factor.}
  \item{stretch}{.}
  \item{correction}{.}
  \item{GSB.adjust}{Logical value. If \code{TRUE}, probe effects in specific binding will
    be adjusted.}
  \item{rho}{correlation coefficient of log background intensity in a pair of pm/mm probes. Default=.7}
  \item{optical.correct}{Logical value. If \code{TRUE}, optical
    background correction is performed.}
  \item{verbose}{Logical value. If \code{TRUE} messages about the progress of
    the function is printed.} 
  \item{fast}{Logical value. If \code{TRUE} a faster ad hoc algorithm is
    used.}
  \item{subset}{a character vector with the the names of the probesets to be
          used in expression calculation.}
  \item{normalize}{logical value. If 'TRUE' normalize data using quantile
    normalization.}
   \item{\dots}{further arguments to be passed (not currently implemented -
          stub for future use).}
}
\details{
  Note that this expression measure is given to you in log base 2
  scale. This differs from most of the other expression measure
  methods.

  The tuning factor \code{k} will have different meanings if one uses
  the fast (add-hoc) algorithm or the empirical Bayes approach. See Wu
  et al. (2003)    
}
\value{
 An \code{ExpressionSet}.
}
\author{Rafeal Irizarry}
\examples{
if(require(affydata) & require(hgu95av2probe) & require(hgu95av2cdf)){
     data(Dilution)
     ai <- compute.affinities(cdfName(Dilution))
     Dil.expr<-gcrma(Dilution,affinity.info=ai,type="affinities")
}
}
\keyword{manip}
